module transactions/locking

open util/ordering[Variable]

sig Transaction {
	holds: set Lock,
	reads: set Variable,
	writes: set Variable
}
abstract sig Lock {
	guards: some Variable
} {
	-- locks are dynamically generated on a per-request basis and destroyed upon release
	--   so each lock that exists is held by some transaction and there are no unheld locks
	Lock = Transaction.holds
}
sig ExclusiveLock, SharedLock extends Lock {}
sig Variable {}

-- returns set of shared locks held by t
fun sharedLocks(t: Transaction): set Lock {
	SharedLock <: (t.holds)
}
-- returns set of exclusive locks held by t
fun exclusiveLocks(t: Transaction): set Lock {
	ExclusiveLock <: (t.holds)
}

fact lockingRules {
	-- each transaction only holds at most one lock for a given variable
	all t: Transaction, disj lock, lock': t.holds | no lock.guards & lock'.guards

	-- if a transaction holds an exclusive lock on some variable, 
	--    no other transaction can acquire locks on same variable
	all t: Transaction, e: exclusiveLocks(t), v: e.guards |
		no t': Transaction - t | v in t'.holds.guards
}

/*
fact noExcessiveLocking {
	-- don't acquire an exclusive lock on a variable if you don't need to write to it
	all t: Transaction, v: Variable | v not in t.writes => v not in exclusiveLocks(t).guards

	-- don't acquire a lock on a variable if you don't need to read or write to it
	all t: Transaction, v: Variable | v not in t.(reads + writes) => v not in t.holds.guards
}
*/

fact noExcessiveLocking {
	all t: Transaction |
		-- don't acquire a shared lock if you don't need to read it any of the variables it guards
		(all lock: sharedLocks(t) | some lock.guards & t.reads) and
		-- don't acquire an exclusive lock if you don't need to write to any of the variables it guards
		(all lock: exclusiveLocks(t) | some lock.guards & t.writes)
}

pred acquireLocksForReads (t: Transaction) {
	-- for all variables that t reads, it must hold a lock for that variable
	all v: t.reads | some lock: t.holds | v in lock.guards
}
pred acquireExclusiveLocksForWrites (t: Transaction) {
	-- for all variables that t writes, it must hold an exclusive lock for that variable
	all v: t.writes | some lock: exclusiveLocks(t) | v in lock.guards
}
pred showAllRequiredLocksAcquired () {
	-- all transactions have all required locks
	all t: Transaction | acquireLocksForReads(t)
	all t: Transaction | acquireExclusiveLocksForWrites(t)

	-- guide the simulation
	some disj t, t': Transaction | some t.reads and some t'.reads
	some writes
}
run showAllRequiredLocksAcquired

pred waitingForReadLock (t: Transaction) {
	-- some other transaction is holding an exclusive lock on a variable 
	--   that t needs to read
	some v: t.reads |
		v not in t.holds.guards and
		some t': Transaction - t, e: exclusiveLocks(t) | v in e.guards
}
pred waitingForWriteLock(t: Transaction) {
	-- some other transaction is holding a lock (either shared or exclusive) 
	--  on a variable that t needs to write to
	some v: t.writes | 
		v not in exclusiveLocks(t).guards and 
		some t': Transaction - t | v in t'.holds.guards
}
pred waitingForLock(t: Transaction) {
	waitingForReadLock(t) or waitingForWriteLock(t)
}
run waitingForLock for 2

pred deadlock() {
	all t: Transaction | waitingForLock(t)
	some ExclusiveLock
}
run deadlock for 2

pred brokenOrderedLockingProtocol () {
	-- don't lock a variable if there's some lesser ranked variable that:
	--   you need to read and haven't obtained a lock for or
	--   you need to write and haven't obtained an exclusive shared lock for
	all t: Transaction | 
		no v: Variable |
			let alreadyLocked = t.holds.guards |
				(some v': alreadyLocked | lt (v, v') and v in t.(reads - holds.guards)) or
				(some v': alreadyLocked | lt (v, v') and v in t.writes - exclusiveLocks(t).guards)
}
assert brokenOrderedLockingProtocolWorks {
	brokenOrderedLockingProtocol() => !deadlock()
}
check brokenOrderedLockingProtocolWorks for 2


pred orderedLockingProtocol () {
	brokenOrderedLockingProtocol ()
	acquireLockOncePerVariable()
	-- either minimal granularity or minimal locking is required for protocol to work
	minimalGranularity() or minimalLocking()
}
pred acquireLockOncePerVariable() {
	-- if you need to read and write a variable, obtain an exclusive lock and not a shared lock
	all t: Transaction, v: t.(writes & reads) | no lock: sharedLocks(t) | v in lock.guards
}
pred minimalGranularity () {
	-- locks are on the same granularity as variables
	all lock: Lock | lone lock.guards
}
pred minimalLocking () {
	all t: Transaction {
	 	all e: exclusiveLocks(t).guards | e in t.writes
		all s: sharedLocks(t).guards | s in t.reads
	}
}
assert orderedLockingWorks {
	orderedLockingProtocol () => !deadlock()
}
check orderedLockingWorks for 5

